# Copyright 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterator, Any, List, Tuple, Sequence, Union, Text, Optional, overload
import socket

import spead2
from spead2 import _PybindStr

class RawItem(object):
    @property
    def id(self) -> int: ...
    @property
    def is_immediate(self) -> bool: ...
    @property
    def immediate_value(self) -> int: ...

class _HeapBase(object):
    @property
    def cnt(self) -> int: ...
    @property
    def flavour(self) -> spead2.Flavour: ...
    def get_items(self) -> List[RawItem]: ...
    def is_start_of_stream(self) -> bool: ...
    def is_end_of_stream(self) -> bool: ...

class Heap(_HeapBase):
    def get_descriptors(self) -> List[spead2.RawDescriptor]: ...

class IncompleteHeap(_HeapBase):
    @property
    def heap_length(self) -> int: ...
    @property
    def received_length(self) -> int: ...
    @property
    def payload_ranges(self) -> List[Tuple[int, int]]: ...

class StreamStats(object):
    heaps: int
    incomplete_heaps_evicted: int
    incomplete_Heaps_flushed: int
    packets: int
    batches: int
    worker_blocked: int
    max_batch: int
    single_packet_heaps: int
    search_dist: int
    def __add__(self, other: StreamStats) -> StreamStats: ...
    def __iadd__(self, other: StreamStats) -> None: ...

class Ringbuffer(object):
    def size(self) -> int: ...
    def capacity(self) -> int: ...

# We make a dummy _Stream base class because mypy objects to the async stream
# type overloading get as a coroutine.
class _Stream(object):
    DEFAULT_UDP_IBV_MAX_SIZE: int = ...
    DEFAULT_UDP_IBV_BUFFER_SIZE: int = ...
    DEFAULT_UDP_IBV_MAX_POLL: int = ...
    DEFAULT_MAX_HEAPS: int = ...
    DEFAULT_RING_HEAPS: int = ...
    DEFAULT_UDP_MAX_SIZE: int = ...
    DEFAULT_UDP_BUFFER_SIZE: int = ...
    DEFAULT_TCP_MAX_SIZE: int = ...
    DEFAULT_TCP_BUFFER_SIZE: int = ...

    def __init__(self, thread_pool: spead2.ThreadPool, bug_compat: int = ...,
                 max_heaps: int = ..., ring_heaps: int = ...,
                 contiguous_only: bool = ...,
                 incomplete_keep_payload_ranges: bool = ...) -> None: ...
    def __iter__(self) -> Iterator[Heap]: ...
    def get_nowait(self) -> Heap: ...
    def set_memory_allocator(self, allocator: spead2.MemoryAllocator) -> None: ...
    def set_memory_pool(self, pool: spead2.MemoryPool) -> None: ...
    def set_memcpy(self, id: int) -> None: ...
    @property
    def stop_on_stop_item(self) -> bool: ...
    @stop_on_stop_item.setter
    def stop_on_stop_item(self, value: bool) -> None: ...
    @property
    def allow_unsized_heaps(self) -> bool: ...
    @allow_unsized_heaps.setter
    def allow_unsized_heaps(self, value: bool) -> None: ...
    def add_buffer_reader(self, buffer: Any) -> None: ...
    @overload
    def add_udp_reader(self, port: int, max_size: int = ..., buffer_size: int = ...,
                       bind_hostname: _PybindStr = ..., socket: Optional[socket.socket] = None) -> None: ...
    @overload
    def add_udp_reader(self, socket: socket.socket, max_size: int = ...) -> None: ...
    @overload
    def add_udp_reader(self, multicast_group: _PybindStr, port: int, max_size: int = ...,
                       buffer_size: int = ..., interface_address: _PybindStr = ...) -> None: ...
    @overload
    def add_udp_reader(self, multicast_group: _PybindStr, port: int, max_size: int = ...,
                       buffer_size: int = ..., interface_index: int = ...) -> None: ...
    @overload
    def add_tcp_reader(self, port: int, max_size: int = ..., buffer_size: int = ...,
                       bind_hostname: _PybindStr = ...) -> None: ...
    @overload
    def add_tcp_reader(self, acceptor: socket.socket, max_size: int = ...) -> None: ...
    @overload
    def add_udp_ibv_reader(self, multicast_group: _PybindStr, port: int,
                           interface_address: _PybindStr,
                           max_size: int = ..., buffer_size: int = ...,
                           comp_vector: int = ..., max_poll: int = ...) -> None: ...
    @overload
    def add_udp_ibv_reader(self, endpoints: Sequence[Tuple[_PybindStr, int]],
                           interface_address: _PybindStr,
                           max_size: int = ..., buffer_size: int = ...,
                           comp_vector: int = ..., max_poll: int = ...) -> None: ...
    def add_udp_pcap_file_reader(self, filename: _PybindStr) -> None: ...
    def add_inproc_reader(self, queue: spead2.InprocQueue) -> None: ...
    def stop(self) -> None: ...
    @property
    def fd(self) -> int: ...
    @property
    def stats(self) -> StreamStats: ...
    @property
    def ringbuffer(self) -> Ringbuffer: ...

class Stream(_Stream):
    def get(self) -> Heap: ...



# Copyright 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import AsyncIterator, Optional

import spead2
import spead2.recv

class Stream(spead2.recv._Stream):
    def __init__(self, thread_pool: spead2.ThreadPool, bug_compat: int = ...,
                 max_heaps: int = ..., ring_heaps: int = ...,
                 contiguous_only: bool = ...,
                 incomplete_keep_payload_ranges: bool = ...,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    async def get(self, loop: Optional[asyncio.AbstractEventLoop] = None) -> spead2.recv.Heap: ...
    def __aiter__(self) -> AsyncIterator[spead2.recv.Heap]: ...



# Generated by bootstrap.sh from spead2/tools/recv_asyncio.py. Do not edit.

# Copyright 2015, 2017-2018 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Receive SPEAD packets and log the contents.

This is both a tool for debugging SPEAD data flows and a demonstrator for the
spead2 package. It thus has many more command-line options than are strictly
necessary, to allow multiple code-paths to be exercised.
"""

from __future__ import print_function, division
import logging
import argparse
import signal

import asyncio


import spead2
import spead2.recv
import spead2.recv.asyncio


def get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('source', nargs='+', help='Sources (filename, host:port or port)')

    group = parser.add_argument_group('Output options')
    group.add_argument('--log', metavar='LEVEL', default='INFO', help='Log level [%(default)s]')
    group.add_argument('--values', action='store_true', help='Show heap values')
    group.add_argument('--descriptors', action='store_true', help='Show descriptors')
    group.add_argument('--max-heaps', type=int, help='Stop receiving after this many heaps')

    group = parser.add_argument_group('Protocol options')
    group.add_argument('--tcp', action='store_true', help='Receive data over TCP instead of UDP')
    group.add_argument('--bind', type=str, default='', help='Interface address for multicast')
    group.add_argument('--pyspead', action='store_true', help='Be bug-compatible with PySPEAD')
    group.add_argument('--joint', action='store_true', help='Treat all sources as a single stream')
    group.add_argument('--packet', type=int, default=spead2.recv.Stream.DEFAULT_UDP_MAX_SIZE,
                       help='Maximum packet size to accept for UDP [%(default)s]')

    group = parser.add_argument_group('Performance options')
    group.add_argument('--buffer', type=int, help='Socket buffer size')
    group.add_argument('--threads', type=int, default=1,
                       help='Number of worker threads [%(default)s]')
    group.add_argument('--heaps', type=int, default=spead2.recv.Stream.DEFAULT_MAX_HEAPS,
                       help='Maximum number of in-flight heaps [%(default)s]')
    group.add_argument('--ring-heaps', type=int, default=spead2.recv.Stream.DEFAULT_RING_HEAPS,
                       help='Ring buffer capacity in heaps [%(default)s]')
    group.add_argument('--mem-pool', action='store_true', help='Use a memory pool')
    group.add_argument('--mem-lower', type=int, default=16384,
                       help='Minimum allocation which will use the memory pool [%(default)s]')
    group.add_argument('--mem-upper', type=int, default=32 * 1024**2,
                       help='Maximum allocation which will use the memory pool [%(default)s]')
    group.add_argument('--mem-max-free', type=int, default=12,
                       help='Maximum free memory buffers [%(default)s]')
    group.add_argument('--mem-initial', type=int, default=8,
                       help='Initial free memory buffers [%(default)s]')
    group.add_argument('--memcpy-nt', action='store_true',
                       help='Use non-temporal memcpy')
    group.add_argument('--affinity', type=spead2.parse_range_list,
                       help='List of CPUs to pin threads to [no affinity]')
    if hasattr(spead2.recv.Stream, 'add_udp_ibv_reader'):
        group.add_argument('--ibv', action='store_true', help='Use ibverbs [no]')
        group.add_argument('--ibv-vector', type=int, default=0, metavar='N',
                           help='Completion vector, or -1 to use polling [%(default)s]')
        group.add_argument('--ibv-max-poll', type=int,
                           default=spead2.recv.Stream.DEFAULT_UDP_IBV_MAX_POLL,
                           help='Maximum number of times to poll in a row [%(default)s]')

    args = parser.parse_args()
    if args.ibv and not args.bind:
        parser.error('--ibv requires --bind')
    if args.tcp and args.ibv:
        parser.error('--ibv and --tcp are incompatible')
    if args.buffer is None:
        if args.tcp:
            args.buffer = spead2.recv.asyncio.Stream.DEFAULT_TCP_BUFFER_SIZE
        else:
            args.buffer = spead2.recv.asyncio.Stream.DEFAULT_UDP_BUFFER_SIZE
    return args


@asyncio.coroutine
def run_stream(stream, name, args):
    try:
        item_group = spead2.ItemGroup()
        num_heaps = 0
        while True:
            try:
                if num_heaps == args.max_heaps:
                    break
                heap = yield from(stream.get())
                print("Received heap {} on stream {}".format(heap.cnt, name))
                num_heaps += 1
                try:
                    if args.descriptors:
                        for raw_descriptor in heap.get_descriptors():
                            descriptor = spead2.Descriptor.from_raw(raw_descriptor, heap.flavour)
                            print('''\
    Descriptor for {0.name} ({0.id:#x})
      description: {0.description}
      format:      {0.format}
      dtype:       {0.dtype}
      shape:       {0.shape}'''.format(descriptor))
                    changed = item_group.update(heap)
                    for (key, item) in changed.items():
                        if args.values:
                            print(key, '=', item.value)
                        else:
                            print(key)
                except ValueError as e:
                    print("Error raised processing heap: {}".format(e))
            except (spead2.Stopped, asyncio.CancelledError):
                print("Shutting down stream {} after {} heaps".format(name, num_heaps))
                stats = stream.stats
                for key in dir(stats):
                    if not key.startswith('_'):
                        print("{}: {}".format(key, getattr(stats, key)))
                break
    finally:
        stream.stop()


def main():
    def make_stream(sources):
        bug_compat = spead2.BUG_COMPAT_PYSPEAD_0_5_2 if args.pyspead else 0
        stream = spead2.recv.asyncio.Stream(thread_pool, bug_compat, args.heaps, args.ring_heaps)
        if memory_pool is not None:
            stream.set_memory_allocator(memory_pool)
        if args.memcpy_nt:
            stream.set_memcpy(spead2.MEMCPY_NONTEMPORAL)
        ibv_endpoints = []
        for source in sources:
            try:
                if ':' in source:
                    host, port = source.rsplit(':', 1)
                    port = int(port)
                else:
                    host = ''
                    port = int(source)
            except ValueError:
                try:
                    stream.add_udp_pcap_file_reader(source)
                except AttributeError:
                    raise RuntimeError('spead2 was compiled without pcap support')
            else:
                if args.tcp:
                    stream.add_tcp_reader(port, args.packet, args.buffer, host)
                elif 'ibv' in args and args.ibv:
                    if host is None:
                        raise ValueError('a multicast group is required when using --ibv')
                    ibv_endpoints.append((host, port))
                elif args.bind and host:
                    stream.add_udp_reader(host, port, args.packet, args.buffer, args.bind)
                else:
                    stream.add_udp_reader(port, args.packet, args.buffer, host)
        if ibv_endpoints:
            stream.add_udp_ibv_reader(ibv_endpoints, args.bind, args.packet,
                                      args.buffer, args.ibv_vector, args.ibv_max_poll)
        return stream

    def make_coro(sources):
        stream = make_stream(sources)
        return run_stream(stream, sources[0], args), stream

    def stop_streams():
        for stream in streams:
            stream.stop()

    args = get_args()
    logging.basicConfig(level=getattr(logging, args.log.upper()))

    if args.affinity is not None and len(args.affinity) > 0:
        spead2.ThreadPool.set_affinity(args.affinity[0])
        thread_pool = spead2.ThreadPool(args.threads, args.affinity[1:] + args.affinity[:1])
    else:
        thread_pool = spead2.ThreadPool(args.threads)
    memory_pool = None
    if args.mem_pool:
        memory_pool = spead2.MemoryPool(args.mem_lower, args.mem_upper,
                                        args.mem_max_free, args.mem_initial)
    if args.joint:
        coros_and_streams = [make_coro(args.source)]
    else:
        coros_and_streams = [make_coro([source]) for source in args.source]
    coros, streams = zip(*coros_and_streams)
    main_task = asyncio.ensure_future(asyncio.gather(*coros))
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(signal.SIGINT, stop_streams)
    try:
        loop.run_until_complete(main_task)
    except asyncio.CancelledError:
        pass
    loop.close()
# Monkey-patch asyncio.ensure_future for Python < 3.4.4.
if not hasattr(asyncio, 'ensure_future'):
    asyncio.ensure_future = getattr(asyncio, 'async')



# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function

import trollius
from trollius import From
import numpy as np
from nose.tools import assert_greater, assert_equal, assert_raises

import spead2
import spead2.send
import spead2.send.trollius
from spead2.send.trollius import UdpStream


class TestUdpStream(object):
    def setup(self):
        # Make a stream slow enough that we can test async interactions
        config = spead2.send.StreamConfig(rate=5e6)
        self.stream = UdpStream(spead2.ThreadPool(), 'localhost', 8888, config)
        self.ig = spead2.send.ItemGroup()
        self.ig.add_item(0x1000, 'test', 'Test item', shape=(256 * 1024,), dtype=np.uint8)
        self.ig['test'].value = np.zeros((256 * 1024,), np.uint8)
        self.heap = self.ig.get_heap()

    @trollius.coroutine
    def _test_async_flush(self):
        assert_greater(self.stream._active, 0)
        yield From(self.stream.async_flush())
        assert_equal(self.stream._active, 0)

    def test_async_flush(self):
        for i in range(3):
            trollius.ensure_future(self.stream.async_send_heap(self.heap))
        # The above only queues up the async sends on the event loop. The rest of the
        # test needs to be run from inside the event loop
        trollius.get_event_loop().run_until_complete(self._test_async_flush())

    def test_async_flush_fail(self):
        """Test async_flush in the case that the last heap sent failed.
        This is arranged by filling up the queue slots first.
        """
        for i in range(5):
            trollius.ensure_future(self.stream.async_send_heap(self.heap))
        # The above only queues up the async sends on the event loop. The rest of the
        # test needs to be run from inside the event loop
        trollius.get_event_loop().run_until_complete(self._test_async_flush())

    def _test_send_error(self, future):
        with assert_raises(IOError):
            yield From(future)

    def test_send_error(self):
        """An error in sending must be reported through the future."""
        # Create a stream with a packet size that is bigger than the likely
        # MTU. It should cause an error.
        stream = UdpStream(
            spead2.ThreadPool(), "localhost", 8888,
            spead2.send.StreamConfig(max_packet_size=100000), buffer_size=0)
        future = stream.async_send_heap(self.heap)
        trollius.get_event_loop().run_until_complete(self._test_send_error(future))


class TestTcpStream(object):
    def _test_connect_failed(self):
        thread_pool = spead2.ThreadPool()
        with assert_raises(IOError):
            yield From(spead2.send.trollius.TcpStream.connect(thread_pool, '127.0.0.1', 8887))

    def test_connect_failed(self):
        trollius.get_event_loop().run_until_complete(self._test_connect_failed())



# Copyright 2015, 2018 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Integration between spead2.recv and trollius
"""
from __future__ import absolute_import
import collections
import functools
import sys

import trollius
from trollius import From, Return

import spead2.recv


# Decorator from official Python documentation to have compatibility with
# both 3.5.0/1 (which expected __aiter__ to return an awaitable) and 3.5.2+
# (which expects it to return an async iterator), and modified to use legal
# Python 2 syntax.
if sys.version_info < (3, 5, 2):
    def _aiter_compat(func):
        @functools.wraps(func)
        @trollius.coroutine
        def wrapper(self):
            raise Return(func(self))
        return wrapper
else:
    def _aiter_compat(func):
        return func


class Stream(spead2.recv.Stream):
    """Stream where `get` is a coroutine that yields the next heap.

    Internally, it maintains a queue of waiters, each represented by a future.
    When a heap becomes available, it is passed to the first waiter. We use
    a callback on a file descriptor being readable, which happens when there
    might be data available. The callback is enabled when we have at least one
    waiter, otherwise disabled.

    The futures store a singleton list containing the heap rather than the heap
    itself. This allows the reference to the heap to be explicitly cleared so
    that the heap can be garbage collected sooner.

    Parameters
    ----------
    loop : event loop, optional
        Default event loop
    """

    def __init__(self, *args, **kwargs):
        self._loop = kwargs.pop('loop', None)
        if self._loop is None:
            self._loop = trollius.get_event_loop()
        super(Stream, self).__init__(*args, **kwargs)
        self._waiters = collections.deque()
        self._listening = False

    def _start_listening(self):
        if not self._listening:
            self._loop.add_reader(self.fd, self._ready_callback)
            self._listening = True

    def _stop_listening(self):
        if self._listening:
            self._loop.remove_reader(self.fd)
            self._listening = False

    def _clear_done_waiters(self):
        """Remove waiters that are done (should only happen if they are cancelled)"""
        while self._waiters and self._waiters[0].done():
            self._waiters.popleft()
        if not self._waiters:
            self._stop_listening()

    def _ready_callback(self):
        self._clear_done_waiters()
        if self._waiters:
            try:
                heap = self.get_nowait()
            except spead2.Empty:
                # Shouldn't happen, but poll may have been woken spuriously
                pass
            except spead2.Stopped as e:
                for waiter in self._waiters:
                    waiter.set_exception(e)
                self._waiters = []
                self._stop_listening()
            else:
                waiter = self._waiters.popleft()
                waiter.set_result([heap])
                if not self._waiters:
                    self._stop_listening()
        # Break cyclic references if spead2.Stopped is raised
        self = None
        waiter = None

    @trollius.coroutine
    def get(self, loop=None):
        """Coroutine that waits for a heap to become available and returns it."""
        self._clear_done_waiters()
        if not self._waiters:
            # If something is available directly, we can avoid going back to
            # the scheduler
            try:
                heap = self.get_nowait()
            except spead2.Empty:
                pass
            else:
                # Give the event loop a chance to run. This ensures that a
                # heap-processing loop cannot live-lock the event loop.
                yield
                raise Return(heap)

        if loop is None:
            loop = self._loop
        waiter = trollius.Future(loop=loop)
        self._waiters.append(waiter)
        self._start_listening()
        heap = (yield From(waiter)).pop()
        raise Return(heap)

    # Asynchronous iterator support for Python 3.5+. It's not supported with
    # trollius, but after passing through trollius2asyncio it becomes useful.
    @_aiter_compat
    def __aiter__(self):
        return self

    @trollius.coroutine
    def __anext__(self):
        try:
            heap = yield From(self.get())
        except spead2.Stopped:
            raise StopAsyncIteration       # noqa: F821 (for Python 2)
        else:
            raise Return(heap)



# Copyright 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Text, Union, Iterator, Optional, overload
import socket

import spead2
from spead2 import _PybindStr

class Heap(object):
    def __init__(self, flavour: spead2.Flavour) -> None: ...
    @property
    def flavour(self) -> spead2.Flavour: ...
    @property
    def repeat_pointers(self) -> bool: ...
    @repeat_pointers.setter
    def repeat_pointers(self, value: bool) -> None: ...
    def add_item(self, item: spead2.Item) -> None: ...
    def add_descriptor(self, descriptor: spead2.Descriptor) -> None: ...
    def add_start(self) -> None: ...
    def add_end(self) -> None: ...

class PacketGenerator(object):
    def __init__(self, heap: Heap, cnt: int, max_packet_size: int) -> None: ...
    def __iter__(self) -> Iterator[bytes]: ...

class StreamConfig(object):
    DEFAULT_MAX_PACKET_SIZE: int = ...
    DEFAULT_MAX_HEAPS: int = ...
    DEFAULT_BURST_SIZE: int = ...
    DEFAULT_BURST_RATE_RATIO: float = ...

    def __init__(self, max_packet_size: int = ..., rate: float = ...,
                 burst_size: int = ..., max_heaps: int = ...,
                 burst_rate_ratio: float = ...) -> None: ...

    @property
    def max_packet_size(self) -> int: ...
    @max_packet_size.setter
    def max_packet_size(self, value: int) -> None: ...

    @property
    def rate(self) -> float: ...
    @rate.setter
    def rate(self, value: float) -> None: ...

    @property
    def burst_size(self) -> int: ...
    @burst_size.setter
    def burst_size(self, value: int) -> None: ...

    @property
    def max_heaps(self) -> int: ...
    @max_heaps.setter
    def max_heaps(self, value: int) -> None: ...

    @property
    def burst_rate_ratio(self) -> float: ...
    @burst_rate_ratio.setter
    def burst_rate_ratio(self, float) -> None: ...

    @property
    def burst_rate(self) -> float: ...

class _Stream(object):
    def set_cnt_sequence(self, next: int, step: int) -> None: ...

class _SyncStream(_Stream):
    def send_heap(self, heap: Heap, cnt: int = ...) -> None: ...

class _UdpStream(object):
    DEFAULT_BUFFER_SIZE: int = ...

class UdpStream(_UdpStream, _SyncStream):
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: StreamConfig,
                 buffer_size: int, socket: socket.socket) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: StreamConfig = ...,
                 buffer_size: int = ..., interface_address: _PybindStr = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: StreamConfig,
                 ttl: int) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: StreamConfig,
                 ttl: int, interface_address: _PybindStr) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: StreamConfig,
                 ttl: int, interface_index: int) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 socket: socket.socket, hostname: _PybindStr, port: int,
                 config: StreamConfig = ...) -> None: ...

class _UdpIbvStream(object):
    DEFAULT_BUFFER_SIZE: int = ...
    DEFAULT_MAX_POLL: int = ...

class UdpIbvStream(_UdpIbvStream, _SyncStream):
    def __init__(self, thread_pool: spead2.ThreadPool,
                 multicast_group: _PybindStr, port: int,
                 config: StreamConfig,
                 interface_address: _PybindStr,
                 buffer_size: int = ..., ttl: int = ...,
                 comp_vector: int = ..., max_pool: int = ...) -> None: ...

class _TcpStream(object):
    DEFAULT_BUFFER_SIZE: int = ...

class TcpStream(_TcpStream, _SyncStream):
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool, socket: socket.socket,
                 config: StreamConfig = ...) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: StreamConfig = ...,
                 buffer_size: int = ..., interface_address: _PybindStr = ...) -> None: ...

class _BytesStream(object):
    def getvalue(self) -> bytes: ...

class BytesStream(_BytesStream, _SyncStream):
    def __init__(self, thread_pool: spead2.ThreadPool,
                 config: StreamConfig = ...) -> None: ...

class _InprocStream(object):
    @property
    def queue(self) -> spead2.InprocQueue: ...

class InprocStream(_InprocStream, _SyncStream):
    def __init__(self, thread_pool: spead2.ThreadPool, queue: spead2.InprocQueue,
                 config: StreamConfig = ...) -> None: ...

class HeapGenerator(object):
    def __init__(self, item_group: spead2.ItemGroup, descriptor_frequency: Optional[int] = None,
                 flavour: spead2.Flavour = spead2.Flavour()) -> None: ...
    def add_to_heap(self, heap: Heap, descriptors: str = ..., data: str = ...) -> Heap: ...
    def get_heap(self, descriptors: str = ..., data: str = ...) -> Heap: ...
    def get_start(self) -> Heap: ...
    def get_end(self) -> Heap: ...

class ItemGroup(spead2.ItemGroup, HeapGenerator):
    def __init__(self, descriptor_frequency: Optional[int] = None,
                 flavour: spead2.Flavour = ...) -> None: ...



# Copyright 2015, 2017 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import division, print_function
import socket
import struct
import time

import numpy as np
import six
from nose.tools import (
    assert_equal, assert_in, assert_is_instance,
    assert_true, assert_false, assert_raises)
try:
    from nose.tools import assert_logs    # Only available from 3.4
except ImportError:
    assert_logs = None

import spead2
import spead2.recv as recv
import spead2.send as send

from .test_common import assert_equal_typed


class Item(object):
    def __init__(self, id, value, immediate=False, offset=0):
        self.id = id
        self.value = value
        self.immediate = immediate
        self.offset = offset

    def encode(self, heap_address_bits):
        if self.immediate:
            return struct.pack('>Q', (1 << 63) | (self.id << heap_address_bits) | self.value)
        else:
            return struct.pack('>Q', (self.id << heap_address_bits) | self.offset)


class Flavour(object):
    def __init__(self, heap_address_bits, bug_compat=0):
        self.heap_address_bits = heap_address_bits
        self.bug_compat = bug_compat

    def _encode_be(self, size, value):
        """Encodes `value` as big-endian in `size` bytes"""
        assert size <= 8
        packed = struct.pack('>Q', value)
        return packed[8 - size:]

    def make_packet(self, items, payload):
        """Generate data for a packet at a low level. The value of non-immediate
        items are taken from the payload; the values in the objects are ignored.
        """
        data = []
        data.append(struct.pack('>BBBBHH', 0x53, 0x4,
                                (64 - self.heap_address_bits) // 8,
                                self.heap_address_bits // 8, 0,
                                len(items)))
        for item in items:
            data.append(item.encode(self.heap_address_bits))
        data.append(bytes(payload))
        return b''.join(data)

    def make_format(self, format):
        if self.bug_compat & spead2.BUG_COMPAT_DESCRIPTOR_WIDTHS:
            field_size = 7
        else:
            field_size = (64 - self.heap_address_bits) // 8
        data = []
        for field in format:
            assert len(field[0]) == 1
            data.append(field[0].encode('ascii'))
            data.append(self._encode_be(field_size, field[1]))
        return b''.join(data)

    def make_shape(self, shape):
        if self.bug_compat & spead2.BUG_COMPAT_DESCRIPTOR_WIDTHS:
            field_size = 8
        else:
            field_size = self.heap_address_bits // 8 + 1
        if self.bug_compat & spead2.BUG_COMPAT_SHAPE_BIT_1:
            variable_marker = 2
        else:
            variable_marker = 1

        data = []
        for value in shape:
            if value is None:
                data.append(struct.pack('>B', variable_marker))
                data.append(self._encode_be(field_size - 1, 0))
            elif value >= 0:
                data.append(self._encode_be(field_size, value))
            else:
                raise ValueError('Shape must contain non-negative values and None')
        return b''.join(data)

    def make_packet_heap(self, heap_cnt, items):
        """Construct a single-packet heap."""
        payload_size = 0
        for item in items:
            if not item.immediate:
                payload_size += len(bytes(item.value))

        all_items = [
            Item(spead2.HEAP_CNT_ID, heap_cnt, True),
            Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
            Item(spead2.PAYLOAD_LENGTH_ID, payload_size, True),
            Item(spead2.HEAP_LENGTH_ID, payload_size, True)]
        offset = 0
        payload = bytearray(payload_size)
        for item in items:
            if not item.immediate:
                value = bytes(item.value)
                all_items.append(Item(item.id, value, offset=offset))
                payload[offset : offset + len(value)] = value
                offset += len(value)
            else:
                all_items.append(item)
        return self.make_packet(all_items, payload)

    def make_plain_descriptor(self, id, name, description, format, shape):
        if not isinstance(name, bytes):
            name = name.encode('ascii')
        if not isinstance(description, bytes):
            description = description.encode('ascii')
        return Item(spead2.DESCRIPTOR_ID, self.make_packet_heap(
            1,
            [
                Item(spead2.DESCRIPTOR_ID_ID, id, True),
                Item(spead2.DESCRIPTOR_NAME_ID, name),
                Item(spead2.DESCRIPTOR_DESCRIPTION_ID, description),
                Item(spead2.DESCRIPTOR_FORMAT_ID, self.make_format(format)),
                Item(spead2.DESCRIPTOR_SHAPE_ID, self.make_shape(shape))
            ]))

    def make_numpy_descriptor_raw(self, id, name, description, header):
        if not isinstance(name, bytes):
            name = name.encode('ascii')
        if not isinstance(description, bytes):
            description = description.encode('ascii')
        if not isinstance(header, bytes):
            header = header.encode('ascii')
        return Item(spead2.DESCRIPTOR_ID, self.make_packet_heap(
            1,
            [
                Item(spead2.DESCRIPTOR_ID_ID, id, True),
                Item(spead2.DESCRIPTOR_NAME_ID, name),
                Item(spead2.DESCRIPTOR_DESCRIPTION_ID, description),
                Item(spead2.DESCRIPTOR_DTYPE_ID, header)
            ]))

    def make_numpy_descriptor(self, id, name, description, dtype, shape, fortran_order=False):
        header = str({
            'descr': np.lib.format.dtype_to_descr(np.dtype(dtype)),
            'fortran_order': bool(fortran_order),
            'shape': tuple(shape)
        })
        return self.make_numpy_descriptor_raw(id, name, description, header)

    def make_numpy_descriptor_from(self, id, name, description, array):
        if array.flags.c_contiguous:
            fortran_order = False
        elif array.flags.f_contiguous:
            fortran_order = True
        else:
            raise ValueError('Array must be C or Fortran-order contiguous')
        return self.make_numpy_descriptor(
            id, name, description, array.dtype, array.shape, fortran_order)


FLAVOUR = Flavour(48)


class TestDecode(object):
    """Various types of descriptors must be correctly interpreted to decode data"""

    def __init__(self):
        self.flavour = FLAVOUR

    def data_to_heaps(self, data, **kwargs):
        """Take some data and pass it through the receiver to obtain a set of heaps.

        Keyword arguments are passed to the receiver constructor.
        """
        thread_pool = spead2.ThreadPool()
        stop_on_stop_item = kwargs.pop('stop_on_stop_item', None)
        allow_unsized_heaps = kwargs.pop('allow_unsized_heaps', None)
        stream = recv.Stream(thread_pool, self.flavour.bug_compat, **kwargs)
        if stop_on_stop_item is not None:
            stream.stop_on_stop_item = stop_on_stop_item
        if allow_unsized_heaps is not None:
            stream.allow_unsized_heaps = allow_unsized_heaps
        stream.add_buffer_reader(data)
        return list(stream)

    def data_to_ig(self, data):
        """Take some data and pass it through the receiver to obtain a single heap,
        from which the items are extracted.
        """
        heaps = self.data_to_heaps(data)
        assert_equal(1, len(heaps))
        ig = spead2.ItemGroup()
        ig.update(heaps[0])
        for name, item in ig.items():
            assert_equal_typed(name, item.name)
        return ig

    def data_to_item(self, data, expected_id):
        """Take some data and pass it through the receiver to obtain a single heap,
        with a single item, which is returned.
        """
        ig = self.data_to_ig(data)
        assert_equal(1, len(ig))
        assert_in(expected_id, ig)
        return ig[expected_id]

    def test_scalar_int(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_scalar_int', 'a scalar integer', [('i', 32)], []),
                Item(0x1234, struct.pack('>i', -123456789))
            ])
        item = self.data_to_item(packet, 0x1234)
        assert_is_instance(item.value, np.int32)
        assert_equal(-123456789, item.value)

    def test_scalar_int_immediate(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_scalar_int', 'a scalar integer', [('u', 32)], []),
                Item(0x1234, 0x12345678, True)
            ])
        item = self.data_to_item(packet, 0x1234)
        assert isinstance(item.value, np.uint32)
        assert_equal(0x12345678, item.value)

    def test_scalar_int_immediate_fastpath(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_scalar_uint', 'a scalar integer', [('u', 48)], []),
                self.flavour.make_plain_descriptor(
                    0x1235, 'test_scalar_positive', 'a positive scalar integer', [('i', 48)], []),
                self.flavour.make_plain_descriptor(
                    0x1236, 'test_scalar_negative', 'a negative scalar integer', [('i', 48)], []),
                Item(0x1234, 0x9234567890AB, True),
                Item(0x1235, 0x1234567890AB, True),
                Item(0x1236, 0x9234567890AB, True)
            ])
        ig = self.data_to_ig(packet)
        assert_equal(3, len(ig))
        assert_equal(0x9234567890AB, ig[0x1234].value)
        assert_equal(0x1234567890AB, ig[0x1235].value)
        assert_equal(-0x6DCBA9876F55, ig[0x1236].value)

    def test_string(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_string', 'a byte string', [('c', 8)], [None]),
                Item(0x1234, 'Hello world'.encode('ascii'))
            ])
        item = self.data_to_item(packet, 0x1234)
        assert_equal_typed('Hello world', item.value)

    def test_array(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_array', 'an array of floats', [('f', 32)], (3, 2)),
                Item(0x1234, struct.pack('>6f', *np.arange(1.5, 7.5)))
            ])
        item = self.data_to_item(packet, 0x1234)
        expected = np.array([[1.5, 2.5], [3.5, 4.5], [5.5, 6.5]], dtype=np.float32)
        np.testing.assert_equal(expected, item.value)

    def test_array_fields(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_array', 'an array of floats', [('f', 32), ('i', 8)], (3,)),
                Item(0x1234, struct.pack('>fbfbfb', 1.5, 1, 2.5, 2, 4.5, -4))
            ])
        item = self.data_to_item(packet, 0x1234)
        dtype = np.dtype('=f4,i1')
        assert_equal(dtype, item.value.dtype)
        expected = np.array([(1.5, 1), (2.5, 2), (4.5, -4)], dtype=dtype)
        np.testing.assert_equal(expected, item.value)

    def test_array_numpy(self):
        expected = np.array([[1.25, 1.75], [2.25, 2.5]], dtype=np.float64)
        send = expected
        if self.flavour.bug_compat & spead2.BUG_COMPAT_SWAP_ENDIAN:
            send = expected.byteswap()
        else:
            send = expected
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor_from(
                    0x1234, 'test_array_numpy', 'an array of floats', expected),
                Item(0x1234, send.data)
            ])
        item = self.data_to_item(packet, 0x1234)
        assert_equal(expected.dtype, item.value.dtype)
        np.testing.assert_equal(expected, item.value)

    def test_array_numpy_fortran(self):
        expected = np.array([[1.25, 1.75], [2.25, 2.5]], dtype=np.float64).T
        assert_false(expected.flags.c_contiguous)
        send = expected
        if self.flavour.bug_compat & spead2.BUG_COMPAT_SWAP_ENDIAN:
            send = expected.byteswap()
        else:
            send = expected
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor_from(
                    0x1234, 'test_array_numpy_fortran', 'an array of floats', expected),
                Item(0x1234, np.ravel(send, 'K').data)
            ])
        item = self.data_to_item(packet, 0x1234)
        assert_equal(expected.dtype, item.value.dtype)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_uint(self):
        expected = np.array([0xABC, 0xDEF, 0x123])
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_uint', 'an array of 12-bit uints', [('u', 12)], (3,)),
                Item(0x1234, b'\xAB\xCD\xEF\x12\x30')
            ])
        item = self.data_to_item(packet, 0x1234)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_int(self):
        expected = np.array([-1348, -529, 291])
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_uint', 'an array of 12-bit ints', [('i', 12)], (3,)),
                Item(0x1234, b'\xAB\xCD\xEF\x12\x30')
            ])
        item = self.data_to_item(packet, 0x1234)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_types(self):
        expected = np.array([(True, 17, 'y', 1.0), (False, -23, 'n', -1.0)], dtype='O,O,S1,>f4')
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_uint', 'an array with bools, int, chars and floats',
                    [('b', 1), ('i', 7), ('c', 8), ('f', 32)], (2,)),
                Item(0x1234, b'\x91y\x3F\x80\x00\x00' + b'\x69n\xBF\x80\x00\x00')
            ])
        item = self.data_to_item(packet, 0x1234)
        np.testing.assert_equal(expected, item.value)

    def test_fallback_scalar(self):
        expected = 0x1234567890AB
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_fallback_scalar', 'a scalar with unusual type',
                    [('u', 48)], ()),
                Item(0x1234, b'\x12\x34\x56\x78\x90\xAB')
            ])
        item = self.data_to_item(packet, 0x1234)
        assert_equal(expected, item.value)

    def test_duplicates(self):
        payload = bytearray(64)
        payload[:] = range(64)
        packets = [
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[0 : 32]),
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 32, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[32 : 64])
        ]
        heaps = self.data_to_heaps(b''.join(packets))
        assert_equal(1, len(heaps))
        items = heaps[0].get_items()
        assert_equal(2, len(items))
        items.sort(key=lambda item: item.id)
        assert_equal(0x1600, items[0].id)
        assert_true(items[0].is_immediate)
        assert_equal(12345, items[0].immediate_value)
        assert_equal(0x5000, items[1].id)
        assert_false(items[1].is_immediate)
        assert_equal(payload, bytearray(items[1]))

    def test_incomplete_heaps(self):
        payload = bytearray(64)
        payload[:] = range(64)
        packets = [
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 5, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 7, True),
                Item(spead2.HEAP_LENGTH_ID, 96, True),
                Item(0x1600, 12345, True),
                Item(0x5000, 0, False, offset=0)], payload[5 : 12]),
            self.flavour.make_packet([
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 32, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True),
                Item(spead2.HEAP_LENGTH_ID, 96, True)], payload[32 : 64])
        ]
        heaps = self.data_to_heaps(b''.join(packets),
                                   contiguous_only=False,
                                   incomplete_keep_payload_ranges=True)
        assert_equal(1, len(heaps))
        assert_is_instance(heaps[0], recv.IncompleteHeap)
        items = heaps[0].get_items()
        assert_equal(1, len(items))   # Addressed item must be excluded
        assert_equal(0x1600, items[0].id)
        assert_equal(96, heaps[0].heap_length)
        assert_equal(39, heaps[0].received_length)
        assert_equal([(5, 12), (32, 64)], heaps[0].payload_ranges)

    def test_is_start_of_stream(self):
        packet = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_START, immediate=True)])
        heaps = self.data_to_heaps(packet)
        assert_true(heaps[0].is_start_of_stream())

        packet = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_DESCRIPTOR_REISSUE, immediate=True)])
        heaps = self.data_to_heaps(packet)
        assert_false(heaps[0].is_start_of_stream())

    def test_is_end_of_stream(self):
        packet = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_STOP, immediate=True)])
        heaps = self.data_to_heaps(packet, stop_on_stop_item=False)
        assert_true(heaps[0].is_end_of_stream())
        assert_false(heaps[0].is_start_of_stream())

    def test_no_stop_on_stop_item(self):
        packet1 = self.flavour.make_packet_heap(
            1,
            [Item(spead2.STREAM_CTRL_ID, spead2.CTRL_STREAM_STOP, immediate=True)])
        packet2 = self.flavour.make_packet_heap(
            2,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_string', 'a byte string', [('c', 8)], [None]),
                Item(0x1234, 'Hello world'.encode('ascii'))
            ])
        heaps = self.data_to_heaps(packet1 + packet2, stop_on_stop_item=False)
        assert_equal(2, len(heaps))
        ig = spead2.ItemGroup()
        ig.update(heaps[0])
        ig.update(heaps[1])
        assert_equal_typed('Hello world', ig['test_string'].value)

    def test_size_mismatch(self):
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'bad', 'an item with insufficient data', [('u', 32)], (5, 5)),
                Item(0x1234, b'\0' * 99)
            ])
        heaps = self.data_to_heaps(packet)
        assert_equal(1, len(heaps))
        ig = spead2.ItemGroup()
        assert_raises(ValueError, ig.update, heaps[0])

    def test_numpy_object(self):
        """numpy dtypes can contain Python objects (by pointer). These can't be
        used for SPEAD.
        """
        dtype = np.dtype('f4,O')
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_numpy_descriptor(
                    0x1234, 'object', 'an item with object pointers', dtype, (5,)),
                Item(0x1234, b'?' * 100)
            ])
        heaps = self.data_to_heaps(packet)
        assert_equal(1, len(heaps))
        ig = spead2.ItemGroup()
        assert_raises(TypeError, ig.update, heaps[0])

    def test_numpy_malformed(self):
        """Malformed numpy header must raise :py:exc:`ValueError`."""
        def helper(header):
            packet = self.flavour.make_packet_heap(
                1,
                [
                    self.flavour.make_numpy_descriptor_raw(
                        0x1234, 'name', 'description', header)
                ])
            heaps = self.data_to_heaps(packet)
            assert_equal(1, len(heaps))
            ig = spead2.ItemGroup()
            assert_raises(ValueError, ig.update, heaps[0])
        helper("{'descr': 'S1'")   # Syntax error: no closing brace
        helper("123")              # Not a dictionary
        helper("import os")        # Security check
        helper("{'descr': 'S1'}")  # Missing keys
        helper("{'descr': 'S1', 'fortran_order': False, 'shape': (), 'foo': 'bar'}")  # Extra keys
        helper("{'descr': 'S1', 'fortran_order': False, 'shape': (-1,)}")    # Bad shape
        helper("{'descr': 1, 'fortran_order': False, 'shape': ()}")          # Bad descriptor
        helper("{'descr': '+-', 'fortran_order': False, 'shape': ()}")       # Bad descriptor
        helper("{'descr': 'S1', 'fortran_order': 0, 'shape': ()}")           # Bad fortran_order
        helper("{'descr': 'S1', 'fortran_order': False, 'shape': (None,)}")  # Bad shape

    def test_nonascii_value(self):
        """Receiving non-ASCII characters in a c8 string must raise
        :py:exc:`UnicodeDecodeError`."""
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'test_string', 'a byte string', [('c', 8)], [None]),
                Item(0x1234, six.u('\u0200').encode('utf-8'))
            ])
        heaps = self.data_to_heaps(packet)
        ig = spead2.ItemGroup()
        assert_raises(UnicodeDecodeError, ig.update, heaps[0])

    def test_nonascii_name(self):
        """Receiving non-ASCII characters in an item name must raise
        :py:exc:`UnicodeDecodeError`."""
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, b'\xEF', 'a byte string', [('c', 8)], [None])
            ])
        heaps = self.data_to_heaps(packet)
        ig = spead2.ItemGroup()
        assert_raises(UnicodeDecodeError, ig.update, heaps[0])

    def test_nonascii_description(self):
        """Receiving non-ASCII characters in an item description must raise
        :py:exc:`UnicodeDecodeError`."""
        packet = self.flavour.make_packet_heap(
            1,
            [
                self.flavour.make_plain_descriptor(
                    0x1234, 'name', b'\xEF', [('c', 8)], [None])
            ])
        heaps = self.data_to_heaps(packet)
        ig = spead2.ItemGroup()
        assert_raises(UnicodeDecodeError, ig.update, heaps[0])

    def test_no_heap_size(self):
        """Heap consisting of packets with no heap size items must work.
        This also tests mixing the general items in with the packet items.
        """
        payload1 = bytes(np.arange(0, 64, dtype=np.uint8).data)
        payload2 = bytes(np.arange(64, 96, dtype=np.uint8).data)
        packet1 = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(0x1000, None, False, offset=0),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 64, True)
            ], payload1)
        packet2 = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 64, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 32, True)
            ], payload2)
        heaps = self.data_to_heaps(packet1 + packet2)
        assert_equal(1, len(heaps))
        raw_items = heaps[0].get_items()
        assert_equal(1, len(raw_items))
        assert_equal(payload1 + payload2, bytearray(raw_items[0]))

    def test_disallow_unsized_heaps(self):
        """Packets without heap length rejected if disallowed"""
        packet = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(0x1000, None, False, offset=0),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 64, True)
            ], bytes(np.arange(0, 64, dtype=np.uint8).data))
        if assert_logs is not None:
            with assert_logs('spead2', 'INFO') as cm:
                heaps = self.data_to_heaps(packet, allow_unsized_heaps=False)
                # Logging is asynchronous, so we have to give it a bit of time
                time.sleep(0.1)
            assert_equal(cm.output, ['INFO:spead2:packet rejected because it has no HEAP_LEN'])
        else:
            # Python 2 fallback
            heaps = self.data_to_heaps(packet, allow_unsized_heaps=False)
        assert_equal(0, len(heaps))

    def test_bad_offset(self):
        """Heap with out-of-range offset should be dropped"""
        packet = self.flavour.make_packet(
            [
                Item(spead2.HEAP_CNT_ID, 1, True),
                Item(spead2.HEAP_LENGTH_ID, 64, True),
                Item(spead2.PAYLOAD_OFFSET_ID, 0, True),
                Item(spead2.PAYLOAD_LENGTH_ID, 64, True),
                Item(0x1000, None, False, offset=65)
            ], b'\0' * 64)
        heaps = self.data_to_heaps(packet)
        assert_equal(0, len(heaps))


class TestStream(object):
    """Tests for the stream API"""

    def __init__(self):
        self.flavour = FLAVOUR

    def test_max_heaps_zero(self):
        """Constructing a stream with max_heaps=0 raises ValueError"""
        with assert_raises(ValueError):
            spead2.recv.Stream(spead2.ThreadPool(), max_heaps=0)

    def test_full_stop(self):
        """Must be able to stop even if the consumer is not consuming
        anything."""
        thread_pool = spead2.ThreadPool(1)
        sender = send.BytesStream(thread_pool)
        ig = send.ItemGroup()
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        gen = send.HeapGenerator(ig)
        for i in range(10):
            sender.send_heap(gen.get_heap(data='all'))
        receiver = spead2.recv.Stream(thread_pool, ring_heaps=4)
        receiver.add_buffer_reader(sender.getvalue())
        # Wait for the ring buffer to block
        while receiver.stats.worker_blocked == 0:
            time.sleep(0.0)
        # Can't usefully check the stats here, because they're only
        # updated at the end of a batch.
        assert_equal(4, receiver.ringbuffer.capacity())
        assert_equal(4, receiver.ringbuffer.size())

        receiver.stop()            # This unblocks all remaining heaps
        stats = receiver.stats
        assert_equal(10, stats.heaps)
        assert_equal(10, stats.packets)
        assert_equal(0, stats.incomplete_heaps_evicted)
        assert_equal(0, stats.incomplete_heaps_flushed)
        assert_equal(1, stats.worker_blocked)
        assert_equal(4, receiver.ringbuffer.capacity())
        assert_equal(4, receiver.ringbuffer.size())

    def test_no_stop_heap(self):
        """A heap containing a stop is not passed to the ring"""
        thread_pool = spead2.ThreadPool(1)
        sender = send.BytesStream(thread_pool)
        ig = send.ItemGroup()
        data = np.array([[6, 7, 8], [10, 11, 12000]], dtype=np.uint16)
        ig.add_item(id=0x2345, name='name', description='description',
                    shape=data.shape, dtype=data.dtype, value=data)
        gen = send.HeapGenerator(ig)
        sender.send_heap(gen.get_heap(data='all'))
        sender.send_heap(gen.get_end())
        receiver = spead2.recv.Stream(thread_pool)
        receiver.add_buffer_reader(sender.getvalue())
        heaps = list(receiver)
        assert_equal(1, len(heaps))

        stats = receiver.stats
        assert_equal(1, stats.heaps)
        assert_equal(2, stats.packets)
        assert_equal(0, stats.incomplete_heaps_evicted)
        assert_equal(0, stats.incomplete_heaps_flushed)
        assert_equal(0, stats.worker_blocked)


class TestUdpReader(object):
    def test_out_of_range_udp_port(self):
        receiver = spead2.recv.Stream(spead2.ThreadPool())
        assert_raises(TypeError, receiver.add_udp_reader, 100000)

    def test_illegal_udp_port(self):
        receiver = spead2.recv.Stream(spead2.ThreadPool())
        assert_raises(RuntimeError, receiver.add_udp_reader, 22)


class TestTcpReader(object):
    def setup(self):
        self.receiver = spead2.recv.Stream(spead2.ThreadPool())
        recv_sock = socket.socket()
        recv_sock.bind(("127.0.0.1", 0))
        recv_sock.listen(1)
        port = recv_sock.getsockname()[1]
        self.receiver.add_tcp_reader(acceptor=recv_sock)
        recv_sock.close()

        self.send_sock = socket.socket()
        self.send_sock.connect(("127.0.0.1", port))

    def teardown(self):
        self.close()

    def close(self):
        if self.send_sock is not None:
            self.send_sock.close()
            self.send_sock = None

    def data_to_heaps(self, data):
        """Take some data and pass it through the receiver to obtain a set of heaps.

        The send socket is closed.
        """
        self.send_sock.send(data)
        self.close()
        return list(self.receiver)

    def data_to_ig(self, data):
        """Take some data and pass it through the receiver to obtain a single heap,
        from which the items are extracted.
        """
        heaps = self.data_to_heaps(data)
        assert_equal(1, len(heaps))
        ig = spead2.ItemGroup()
        ig.update(heaps[0])
        for name, item in ig.items():
            assert_equal_typed(name, item.name)
        return ig

    def data_to_item(self, data, expected_id):
        """Take some data and pass it through the receiver to obtain a single heap,
        with a single item, which is returned.
        """
        ig = self.data_to_ig(data)
        assert_equal(1, len(ig))
        assert_in(expected_id, ig)
        return ig[expected_id]

    def simple_packet(self):
        return FLAVOUR.make_packet_heap(
            1,
            [
                FLAVOUR.make_plain_descriptor(
                    0x1234, 'test_scalar_int', 'a scalar integer', [('u', 32)], []),
                Item(0x1234, 0x12345678, True)
            ])

    def test_bad_header(self):
        """A nonsense header followed by a normal packet"""
        data = b'deadbeef' + self.simple_packet()
        item = self.data_to_item(data, 0x1234)
        assert_is_instance(item.value, np.uint32)
        assert_equal(0x12345678, item.value)

    def test_packet_too_big(self):
        """A packet that is too large (should be rejected) followed by a normal packet"""
        zeros = np.zeros(100000, np.uint8)
        packet1 = FLAVOUR.make_packet_heap(
            1,
            [
                FLAVOUR.make_numpy_descriptor_from(
                    0x2345, 'test_big_array', 'over-sized packet', zeros),
                Item(0x2345, zeros.data)
            ])
        data = packet1 + self.simple_packet()
        item = self.data_to_item(data, 0x1234)
        assert_is_instance(item.value, np.uint32)
        assert_equal(0x12345678, item.value)

    def test_partial_header(self):
        """Connection closed partway through a header"""
        packet = self.simple_packet()
        heaps = self.data_to_heaps(packet[:6])
        assert_equal([], heaps)

    def test_partial_packet(self):
        """Connection closed partway through item descriptors"""
        packet = self.simple_packet()
        heaps = self.data_to_heaps(packet[:45])
        assert_equal([], heaps)

    def test_partial_payload(self):
        """Connection closed partway through item payload"""
        packet = self.simple_packet()
        heaps = self.data_to_heaps(packet[:-1])
        assert_equal([], heaps)



# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Integration between spead2.send and trollius
"""
from __future__ import absolute_import

import trollius
from trollius import From, Return

from spead2._spead2.send import UdpStreamAsyncio as _UdpStreamAsyncio
from spead2._spead2.send import TcpStreamAsyncio as _TcpStreamAsyncio
from spead2._spead2.send import InprocStreamAsyncio as _InprocStreamAsyncio


def _wrap_class(name, base_class):
    class Wrapped(base_class):
        def __init__(self, *args, **kwargs):
            self._loop = kwargs.pop('loop', None)
            super(Wrapped, self).__init__(*args, **kwargs)
            if self._loop is None:
                self._loop = trollius.get_event_loop()
            self._active = 0
            self._last_queued_future = None

        def async_send_heap(self, heap, cnt=-1, loop=None):
            """Send a heap asynchronously. Note that this is *not* a coroutine:
            it returns a future. Adding the heap to the queue is done
            synchronously, to ensure proper ordering.

            Parameters
            ----------
            heap : :py:class:`spead2.send.Heap`
                Heap to send
            cnt : int, optional
                Heap cnt to send (defaults to auto-incrementing)
            loop : :py:class:`trollius.BaseEventLoop`, optional
                Event loop to use, overriding the constructor.
            """

            if loop is None:
                loop = self._loop
            future = trollius.Future(loop=self._loop)

            def callback(exc, bytes_transferred):
                if exc is not None:
                    future.set_exception(exc)
                else:
                    future.set_result(bytes_transferred)
                self._active -= 1
                if self._active == 0:
                    self._loop.remove_reader(self.fd)
                    self._last_queued_future = None  # Purely to free the memory
            queued = super(Wrapped, self).async_send_heap(heap, callback, cnt)
            if self._active == 0:
                self._loop.add_reader(self.fd, self.process_callbacks)
            self._active += 1
            if queued:
                self._last_queued_future = future
            return future

        @trollius.coroutine
        def async_flush(self):
            """Asynchronously wait for all enqueued heaps to be sent. Note that
            this only waits for heaps passed to :meth:`async_send_heap` prior to
            this call, not ones added while waiting."""
            future = self._last_queued_future
            if future is not None:
                yield From(trollius.wait([future]))

    Wrapped.__name__ = name
    return Wrapped


UdpStream = _wrap_class('UdpStream', _UdpStreamAsyncio)
UdpStream.__doc__ = \
    """SPEAD over UDP with asynchronous sends. The other constructors
    defined for :py:class:`spead2.send.UdpStream` are also applicable here.

    Parameters
    ----------
    thread_pool : :py:class:`spead2.ThreadPool`
        Thread pool handling the I/O
    hostname : str
        Peer hostname
    port : int
        Peer port
    config : :py:class:`spead2.send.StreamConfig`
        Stream configuration
    buffer_size : int
        Socket buffer size. A warning is logged if this size cannot be set due
        to OS limits.
    loop : :py:class:`trollius.BaseEventLoop`, optional
        Event loop to use (defaults to ``trollius.get_event_loop()``)
    """

_TcpStreamBase = _wrap_class('TcpStream', _TcpStreamAsyncio)


class TcpStream(_TcpStreamBase):
    """SPEAD over TCP with asynchronous connect and sends.

    Most users will use :py:meth:`connect` to asynchronously create a stream.
    The constructor should only be used if you wish to provide your own socket
    and take care of connecting yourself.

    Parameters
    ----------
    thread_pool : :py:class:`spead2.ThreadPool`
        Thread pool handling the I/O
    socket : :py:class:`socket.socket`
        TCP/IP Socket that is already connected to the remote end
    config : :py:class:`spead2.send.StreamConfig`
        Stream configuration
    """

    @classmethod
    @trollius.coroutine
    def connect(cls, *args, **kwargs):
        """Open a connection.

        The arguments are the same as for the constructor of
        :py:class:`spead2.send.TcpStream`.
        """
        loop = kwargs.get('loop')
        if loop is None:
            loop = trollius.get_event_loop()
        future = trollius.Future(loop=loop)

        def callback(arg):
            if not future.done():
                if isinstance(arg, Exception):
                    loop.call_soon_threadsafe(future.set_exception, arg)
                else:
                    loop.call_soon_threadsafe(future.set_result, arg)

        stream = cls(callback, *args, **kwargs)
        yield From(future)
        raise Return(stream)


InprocStream = _wrap_class('InprocStream', _InprocStreamAsyncio)
InprocStream.__doc__ = \
    """SPEAD over reliable in-process transport.

    .. note::

        Data may still be lost if the maximum number of in-flight heaps (set
        in the stream config) is exceeded. Either set this value to more
        heaps than will ever be sent (which will use unbounded memory) or be
        sure to block on the futures returned before exceeding the capacity.

    Parameters
    ----------
    thread_pool : :py:class:`spead2.ThreadPool`
        Thread pool handling the I/O
    queue : :py:class:`spead2.InprocQueue`
        Queue holding the data in flight
    config : :py:class:`spead2.send.StreamConfig`
        Stream configuration
    """

try:
    from spead2._spead2.send import UdpIbvStreamAsyncio as _UdpIbvStreamAsyncio

    UdpIbvStream = _wrap_class('UdpIbvStream', _UdpIbvStreamAsyncio)
    UdpIbvStream.__doc__ = \
        """Like :class:`UdpStream`, but using the Infiniband Verbs API.

        Parameters
        ----------
        thread_pool : :py:class:`spead2.ThreadPool`
            Thread pool handling the I/O
        multicast_group : str
            IP address (or DNS name) of multicast group to join
        port : int
            Peer port
        config : :py:class:`spead2.send.StreamConfig`
            Stream configuration
        interface_address : str
            IP address of network interface from which to send
        buffer_size : int, optional
            Buffer size
        ttl : int, optional
            Time-To-Live of packets
        comp_vector : int, optional
            Completion channel vector (interrupt)
            for asynchronous operation, or
            a negative value to poll continuously. Polling
            should not be used if there are other users of the
            thread pool. If a non-negative value is provided, it
            is taken modulo the number of available completion
            vectors. This allows a number of readers to be
            assigned sequential completion vectors and have them
            load-balanced, without concern for the number
            available.
        max_poll : int
            Maximum number of times to poll in a row, without
            waiting for an interrupt (if `comp_vector` is
            non-negative) or letting other code run on the
            thread (if `comp_vector` is negative).
        loop : :py:class:`trollius.BaseEventLoop`, optional
            Event loop to use (defaults to ``trollius.get_event_loop()``)
        """

except ImportError:
    pass



#!/usr/bin/env python

# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import spead2
import spead2.send
import spead2.send.trollius
import logging
import numpy as np
import trollius

logging.basicConfig(level=logging.INFO)

thread_pool = spead2.ThreadPool()
stream = spead2.send.trollius.UdpStream(
    thread_pool, "127.0.0.1", 8888, spead2.send.StreamConfig(rate=1e7))
del thread_pool  # Make sure this doesn't crash anything

shape = (40, 50)
ig = spead2.send.ItemGroup(flavour=spead2.Flavour(4, 64, 48, spead2.BUG_COMPAT_PYSPEAD_0_5_2))
item = ig.add_item(0x1234, 'foo', 'a foo item', shape=shape, dtype=np.int32)
item.value = np.zeros(shape, np.int32)
futures = [
    stream.async_send_heap(ig.get_heap()),
    stream.async_send_heap(ig.get_end())
]
# Delete things to check that there are no refcounting bugs
del ig
del stream
trollius.get_event_loop().run_until_complete(trollius.wait(futures))



# Copyright 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import socket
from typing import Optional, overload

import spead2
import spead2.send

from spead2 import _PybindStr

class _AsyncStream(spead2.send._Stream):
    @property
    def fd(self) -> int: ...
    def flush(self) -> None: ...
    def async_send_heap(self, heap: spead2.send.Heap, cnt: int = ...,
                        loop: Optional[asyncio.AbstractEventLoop] = None) -> asyncio.Future[int]: ...
    async def async_flush(self) -> None: ...

class UdpStream(spead2.send._UdpStream, _AsyncStream):
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: spead2.send.StreamConfig,
                 buffer_size: int, socket: socket.socket,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: spead2.send.StreamConfig = ...,
                 buffer_size: int = ..., interface_address: _PybindStr = ...,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: spead2.send.StreamConfig,
                 ttl: int,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: spead2.send.StreamConfig,
                 ttl: int, interface_address: _PybindStr,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 hostname: _PybindStr, port: int,
                 config: spead2.send.StreamConfig,
                 ttl: int, interface_index: int,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    @overload
    def __init__(self, thread_pool: spead2.ThreadPool,
                 socket: socket.socket, hostname: _PybindStr, port: int,
                 config: spead2.send.StreamConfig = ...,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...


class TcpStream(spead2.send._TcpStream, _AsyncStream):
    def __init__(self, thread_pool: spead2.ThreadPool, socket: socket.socket,
                 config: spead2.send.StreamConfig = ...,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...
    @classmethod
    async def connect(self, thread_pool: spead2.ThreadPool,
                      hostname: _PybindStr, port: int,
                      config: spead2.send.StreamConfig = ...,
                      buffer_size: int = ..., interface_address: _PybindStr = ...,
                      *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...


class InprocStream(spead2.send._InprocStream, _AsyncStream):
    def __init__(self, thread_pool: spead2.ThreadPool, queue: spead2.InprocQueue,
                 config: spead2.send.StreamConfig = ...,
                 *, loop: Optional[asyncio.AbstractEventLoop] = None) -> None: ...




"""Tests for shutdown ordering.

These cannot be run through nosetests because they deal with interpreter
shutdown.
"""

import logging

import spead2
import spead2.recv
import spead2.send
import spead2._spead2


def test_logging_shutdown():
    """Spam the log with lots of messages, then terminate.

    The logging thread needs to be gracefully cleaned up.
    """
    # Set a log level that won't actually display the messages.
    logging.basicConfig(level=logging.ERROR)
    for i in range(20000):
        spead2._spead2.log_info('Test message {}'.format(i))


def test_running_thread_pool():
    global tp
    tp = spead2.ThreadPool()


def test_running_stream():
    global stream
    logging.basicConfig(level=logging.ERROR)
    stream = spead2.recv.Stream(spead2.ThreadPool())
    stream.add_udp_reader(7148)
    sender = spead2.send.UdpStream(spead2.ThreadPool(), 'localhost', 7148)
    ig = spead2.send.ItemGroup()
    ig.add_item(id=None, name='test', description='test',
                shape=(), format=[('u', 32)], value=0xdeadbeef)
    heap = ig.get_heap()
    for i in range(5):
        sender.send_heap(heap)



#!/usr/bin/env python

# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Benchmark tool to estimate the sustainable SPEAD bandwidth between two
machines, for a specific set of configurations.

Since UDP is lossy, this is not a trivial problem. We binary search for the
speed that is just sustainable. To make the test at a specific speed more
reliable, it is repeated several times, opening a new stream each time, and
with a delay to allow processors to return to idle states. A TCP control
stream is used to synchronise the two ends. All configuration is done on
the master end.
"""

import sys


if sys.version_info >= (3, 4):
    from spead2.tools import bench_asyncio
    bench_asyncio.main()
else:
    from spead2.tools import bench_trollius
    bench_trollius.main()



# Copyright 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import (List, Sequence, Optional, Tuple, Any, Union,
                    Dict, KeysView, ValuesView, Text, overload)

import numpy as np
import spead2.recv


# Not the same as typing.AnyStr, which is a TypeVar
_PybindStr = Union[Text, bytes]

__version__: str

BUG_COMPAT_DESCRIPTOR_WIDTHS: int
BUG_COMPAT_SHAPE_BIT_1: int
BUG_COMPAT_SWAP_ENDIAN: int
BUG_COMPAT_PYSPEAD_0_5_2: int

NULL_ID: int
HEAP_CNT_ID: int
HEAP_LENGTH_ID: int
PAYLOAD_OFFSET_ID: int
PAYLOAD_LENGTH_ID: int
DESCRIPTOR_ID: int
STREAM_CTRL_ID: int

DESCRIPTOR_NAME_ID: int
DESCRIPTOR_DESCRIPTION_ID: int
DESCRIPTOR_SHAPE_ID: int
DESCRIPTOR_FORMAT_ID: int
DESCTIPTOR_DTYPE_ID: int

CTRL_STREAM_START: int
CTRL_DESCRIPTOR_REISSUE: int
CTRL_STREAM_STOP: int
CTRL_DESCRIPTOR_UPDATE: int

MEMCPY_STD: int
MEMCPY_NONTEMPORAL: int

class Stopped(RuntimeError):
    pass

class Empty(RuntimeError):
    pass

class Flavour(object):
    @overload
    def __init__(self, version: int, item_pointer_bits : int,
                 heap_address_bits: int, bug_compat: int = ...) -> None: ...
    @overload
    def __init__(self) -> None: ...
    def __eq__(self, o: object) -> bool: ...
    def __ne__(self, o: object) -> bool: ...
    @property
    def version(self) -> int: ...
    @property
    def item_pointer_bits(self) -> int: ...
    @property
    def heap_address_bits(self) -> int: ...
    @property
    def bug_compat(self) -> int: ...

class ThreadPool(object):
    @overload
    def __init__(self, threads: int = ...) -> None: ...
    @overload
    def __init__(self, threads: int, affinity: List[int]) -> None: ...

class MemoryAllocator(object):
    def __init__(self) -> None: ...

class MmapAllocator(MemoryAllocator):
    def __init__(self, flags: int = ...) -> None: ...

class MemoryPool(MemoryAllocator):
    @overload
    def __init__(self, lower: int, upper: int, max_free: int, initial: int,
                 allocator: Optional[MemoryAllocator] = None) -> None: ...
    @overload
    def __init__(self, thread_pool: ThreadPool, lower: int, upper: int, max_free: int, initial: int,
                 low_water: int, allocator: MemoryAllocator) -> None: ...
    @property
    def warn_on_empty(self) -> bool: ...
    @warn_on_empty.setter
    def warn_on_empty(self, value: bool) -> None: ...

class InprocQueue(object):
    def __init__(self) -> None: ...
    def stop(self) -> None: ...

class RawDescriptor(object):
    @property
    def id(self) -> int: ...
    @id.setter
    def id(self, value: int) -> None: ...

    @property
    def name(self) -> bytes: ...
    @name.setter
    def name(self, value: bytes) -> None: ...

    @property
    def description(self) -> bytes: ...
    @description.setter
    def description(self, value: bytes) -> None: ...

    @property
    def shape(self) -> List[Optional[int]]: ...
    @shape.setter
    def shape(self, value: Sequence[Optional[int]]) -> None: ...

    @property
    def format(self) -> List[Tuple[_PybindStr, int]]: ...
    @format.setter
    def format(self, value: List[Tuple[_PybindStr, int]]) -> None: ...

    @property
    def numpy_header(self) -> bytes: ...
    @numpy_header.setter
    def numpy_header(self, value: bytes) -> None: ...

class IbvContext(object):
    def __init__(self, interface_address: _PybindStr) -> None: ...
    def reset(self) -> None: ...

def parse_range_list(ranges: str) -> List[int]: ...

class Descriptor(object):
    id: int
    name: str
    description: str
    shape: Sequence[Optional[int]]
    dtype: Optional[np.dtype]
    order: str
    format: Optional[List[Tuple[str, int]]]

    def __init__(self, id: int, name: str, description: str,
                 shape: Sequence[Optional[int]], dtype: Optional[np.dtype] = None,
                 order: str = ..., format: Optional[List[Tuple[str, int]]] = None) -> None: ...
    @property
    def itemsize_bits(self) -> int: ...
    @property
    def is_variable_size(self) -> bool: ...
    @property
    def allow_immediate(self) -> bool: ...
    def dynamic_shape(self, max_elements: int) -> Sequence[int]: ...
    def compatible_shape(self, shape: Sequence[int]) -> bool: ...
    @classmethod
    def from_raw(cls, raw_descriptor: RawDescriptor, flavour: Flavour): ...
    def to_raw(self, flavour: Flavour) -> RawDescriptor: ...

class Item(Descriptor):
    version: int

    def __init__(self, id: int, name: str, description: str,
                 shape: Sequence[Optional[int]], dtype: Optional[np.dtype] = None,
                 order: str = ..., format: Optional[List[Tuple[str, int]]] = None,
                 value: Any = None) -> None: ...
    @property
    def value(self) -> Any: ...
    @value.setter
    def value(self, new_value: Any) -> None: ...
    def set_from_raw(self, raw_item: spead2.recv.RawItem) -> None: ...
    # typing has no buffer protocol ABC (https://bugs.python.org/issue27501)
    def to_buffer(self) -> Any: ...

class ItemGroup(object):
    def __init__(self) -> None: ...
    def add_item(self, id: Optional[int], name: str, description: str,
                 shape: Sequence[Optional[int]], dtype: Optional[np.dtype] = None,
                 order: str = 'C', format: Optional[List[Tuple[str, int]]] = None,
                 value: Any = None) -> Item: ...
    def __getitem__(self, key: Union[int, str]) -> Item: ...
    def __contains__(self, key: Union[int, str]) -> bool: ...
    def keys(self) -> KeysView[str]: ...
    def ids(self) -> KeysView[int]: ...
    def values(self) -> ValuesView[Item]: ...
    def __len__(self) -> int: ...
    def update(self, heap: recv.Heap) -> Dict[str, Item]: ...



# Copyright 2015 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Send SPEAD protocol"""

from __future__ import print_function, division
import weakref

import spead2 as _spead2
from spead2._spead2.send import (       # noqa: F401
    StreamConfig, Heap, PacketGenerator,
    BytesStream, UdpStream, TcpStream, InprocStream)
try:
    from spead2._spead2.send import UdpIbvStream      # noqa: F401
except ImportError:
    pass


class _ItemInfo(object):
    def __init__(self, item):
        self.version = None
        self.descriptor_cnt = None
        self.item = weakref.ref(item)


class HeapGenerator(object):
    """Tracks which items and item values have previously been sent and
    generates delta heaps.

    Parameters
    ----------
    item_group : :py:class:`spead2.ItemGroup`
        Item group to monitor.
    descriptor_frequency : int, optional
        If specified, descriptors will be re-sent once every `descriptor_frequency` heaps
        generated by this method.
    flavour : :py:class:`spead2.Flavour`
        The SPEAD protocol flavour used for heaps generated by :py:meth:`get_heap` and
        :py:meth:`get_end`.
    """
    def __init__(self, item_group, descriptor_frequency=None, flavour=_spead2.Flavour()):
        self._item_group = item_group
        self._info = {}              # Maps ID to _ItemInfo
        self._descriptor_frequency = descriptor_frequency
        # Counter for calls to add_to_heap. This is independent of the
        # protocol-level heap count.
        self._descriptor_cnt = 0
        self._flavour = flavour

    def _get_info(self, item):
        if item.id not in self._info:
            self._info[item.id] = _ItemInfo(item)
        return self._info[item.id]

    def _descriptor_stale(self, item, info):
        if info.descriptor_cnt is None:
            # Never been sent before
            return True
        if self._descriptor_frequency is not None \
                and self._descriptor_cnt - info.descriptor_cnt >= self._descriptor_frequency:
            # This descriptor is due for a resend
            return True
        # Check for complete replacement of the item
        orig_item = info.item()
        if orig_item is not item:
            info.version = None
            info.item = weakref.ref(item)
            return True
        return False

    def add_to_heap(self, heap, descriptors='stale', data='stale'):
        """Update a heap to contains all the new items and item descriptors
        since the last call.

        Parameters
        ----------
        heap : :py:class:`Heap`
            The heap to update.
        descriptors : {'stale', 'all', 'none'}
            Which descriptors to send. The default ('stale') sends only
            descriptors that have not been sent, or have not been sent recently
            enough according to the `descriptor_frequency` passed to the
            constructor. The other options are to send all the descriptors or
            none of them. Sending all descriptors is useful if a new receiver
            is added which will be out of date.
        data : {'stale', 'all', 'none'}
            Which data items to send.
        item_group : :py:class:`ItemGroup`, optional
            If specified, uses the items from this item group instead of the
            one passed to the constructor (which could be `None`).

        Raises
        ------
        ValueError
            if `descriptors` or `data` is not one of the legal values
        """
        if descriptors not in ['stale', 'all', 'none']:
            raise ValueError("descriptors must be one of 'stale', 'all', 'none'")
        if data not in ['stale', 'all', 'none']:
            raise ValueError("data must be one of 'stale', 'all', 'none'")
        for item in self._item_group.values():
            info = self._get_info(item)
            if (descriptors == 'all') or (descriptors == 'stale'
                                          and self._descriptor_stale(item, info)):
                heap.add_descriptor(item)
                info.descriptor_cnt = self._descriptor_cnt
            if item.value is not None:
                if (data == 'all') or (data == 'stale' and info.version != item.version):
                    heap.add_item(item)
                    info.version = item.version
        self._descriptor_cnt += 1
        return heap

    def get_heap(self, *args, **kwargs):
        """Return a new heap which contains all the new items and item
        descriptors since the last call. This is a convenience wrapper
        around :meth:`add_to_heap`.
        """
        heap = Heap(self._flavour)
        self.add_to_heap(heap, *args, **kwargs)
        return heap

    def get_start(self):
        """Return a heap that contains only a start-of-stream marker.
        """
        heap = Heap(self._flavour)
        heap.add_start()
        return heap

    def get_end(self):
        """Return a heap that contains only an end-of-stream marker.
        """
        heap = Heap(self._flavour)
        heap.add_end()
        return heap


class ItemGroup(_spead2.ItemGroup, HeapGenerator):
    """Bundles an ItemGroup and HeapGenerator into a single class"""
    def __init__(self, *args, **kwargs):
        _spead2.ItemGroup.__init__(self)
        HeapGenerator.__init__(self, self, *args, **kwargs)



#!/usr/bin/env python

# Copyright 2015, 2017, 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
import glob
import sys
import os
import os.path
import subprocess


def find_version():
    # Cannot simply import it, since that tries to import spead2 as well, which
    # isn't built yet.
    globals_ = {}
    with open(os.path.join(os.path.dirname(__file__), 'spead2', '_version.py')) as f:
        code = f.read()
    exec(code, globals_)
    return globals_['__version__']


# Restrict installed modules to those appropriate to the Python version
class BuildPy(build_py):
    def find_package_modules(self, package, package_dir):
        # distutils uses old-style classes, so no super
        modules = build_py.find_package_modules(self, package, package_dir)
        if sys.version_info < (3, 4):
            modules = [m for m in modules if not m[1].endswith('asyncio')]
        if sys.version_info < (3, 5):
            modules = [m for m in modules if not m[1].endswith('py35')]
        if sys.version_info >= (3, 7):
            modules = [m for m in modules if not m[1].endswith('trollius')]
        return modules


class BuildExt(build_ext):
    user_options = build_ext.user_options + [
        ('coverage', None,
         "build with GCC --coverage option")
    ]
    boolean_options = build_ext.boolean_options + ['coverage']

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.coverage = None

    def run(self):
        self.mkpath(self.build_temp)
        subprocess.check_call(os.path.abspath('configure'), cwd=self.build_temp)
        # Ugly hack to add libraries conditional on configure result
        have_pcap = False
        with open(os.path.join(self.build_temp, 'include', 'spead2', 'common_features.h')) as f:
            for line in f:
                if line.strip() == '#define SPEAD2_USE_PCAP 1':
                    have_pcap = True
        for extension in self.extensions:
            if have_pcap:
                extension.libraries.extend(['pcap'])
            if self.coverage:
                extension.extra_compile_args.extend(['-g', '--coverage'])
                extension.libraries.extend(['gcov'])
            extension.include_dirs.insert(0, os.path.join(self.build_temp, 'include'))
        # distutils uses old-style classes, so no super
        build_ext.run(self)

    def build_extensions(self):
        # Stop GCC complaining about -Wstrict-prototypes in C++ code
        try:
            self.compiler.compiler_so.remove('-Wstrict-prototypes')
        except ValueError:
            pass
        build_ext.build_extensions(self)


# Can't actually install on readthedocs.org because Boost.Python is missing,
# but we need setup.py to still be successful to make the doc build work.
rtd = os.environ.get('READTHEDOCS') == 'True'

if not rtd:
    if not os.path.exists(os.path.join(os.path.dirname(__file__), 'configure')):
        raise SystemExit("configure not found. Either download a release " +
                         "from https://pypi.python.org/pypi/spead2 or run " +
                         "./bootstrap.sh if not using a release.")
    if not os.path.exists(os.path.join(
            os.path.dirname(__file__),
            '3rdparty', 'pybind11', 'include', 'pybind11', 'pybind11.h')):
        raise SystemExit("pybind11 not found. Either download a release " +
                         "from https://pypi.python.org/pypi/spead2 or run " +
                         "git submodule update --init --recursive if not " +
                         "using a release.")

    libraries = ['boost_system']

    extensions = [
        Extension(
            '_spead2',
            sources=(glob.glob('src/common_*.cpp') +
                     glob.glob('src/recv_*.cpp') +
                     glob.glob('src/send_*.cpp') +
                     glob.glob('src/py_*.cpp')),
            depends=glob.glob('include/spead2/*.h'),
            language='c++',
            include_dirs=['include', '3rdparty/pybind11/include'],
            extra_compile_args=['-std=c++11', '-g0', '-fvisibility=hidden'],
            libraries=libraries)
    ]
else:
    extensions = []

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme_file:
    readme = readme_file.read()

setup(
    author='Bruce Merry',
    author_email='bmerry@ska.ac.za',
    name='spead2',
    version=find_version(),
    description='High-performance SPEAD implementation',
    long_description=readme,
    url='https://github.com/ska-sa/spead2',
    license='LGPLv3+',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Networking'],
    ext_package='spead2',
    ext_modules=extensions,
    cmdclass={'build_ext': BuildExt, 'build_py': BuildPy},
    install_requires=[
        'numpy>=1.9.2',
        'six',
        'trollius; python_version<"3.4"'
    ],
    tests_require=[
        'netifaces',
        'nose',
        'decorator',
        'trollius; python_version<"3.7"',
        'asynctest; python_version>="3.5"'
    ],
    test_suite='nose.collector',
    packages=find_packages(),
    package_data={'': ['py.typed', '*.pyi']},
    scripts=glob.glob('scripts/*.py')
)



#!/usr/bin/env python

# Copyright 2019 SKA South Africa
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import copy

import jinja2
from pycparser import c_ast
from pycparser.c_parser import CParser
from pycparser.c_generator import CGenerator


PREFIX = '''
/* Copyright 2019 SKA South Africa
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file
 *
 * This file is automatically generated. Do not edit.
 */
'''

# Extracted signatures for symbols needed from libibverbs. The typedefs are
# arbitrary and just used to allow pycparser to parse the code.
#
# Note that some functions in infiniband/verbs.h are implemented as static
# inline functions, and so do not get listed here.
INPUT = '''
typedef unsigned long size_t;
typedef unsigned long uint64_t;

void ibv_ack_cq_events(struct ibv_cq *cq, unsigned int nevents);

struct ibv_pd *ibv_alloc_pd(struct ibv_context *context);

int ibv_close_device(struct ibv_context *context);

struct ibv_comp_channel *ibv_create_comp_channel(struct ibv_context *context);

struct ibv_cq *ibv_create_cq(struct ibv_context *context, int cqe,
                             void *cq_context,
                             struct ibv_comp_channel *channel,
                             int comp_vector);

struct ibv_qp *ibv_create_qp(struct ibv_pd *pd,
                             struct ibv_qp_init_attr *qp_init_attr);

int ibv_dealloc_pd(struct ibv_pd *pd);

int ibv_dereg_mr(struct ibv_mr *mr);

int ibv_destroy_comp_channel(struct ibv_comp_channel *channel);

int ibv_destroy_cq(struct ibv_cq *cq);

int ibv_destroy_qp(struct ibv_qp *qp);

void ibv_free_device_list(struct ibv_device **list);

int ibv_get_cq_event(struct ibv_comp_channel *channel,
                     struct ibv_cq **cq, void **cq_context);

uint64_t ibv_get_device_guid(struct ibv_device *device);

struct ibv_device **ibv_get_device_list(int *num_devices);

struct ibv_context *ibv_open_device(struct ibv_device *device);

int ibv_modify_qp(struct ibv_qp *qp, struct ibv_qp_attr *attr,
                  int attr_mask);

int ibv_query_device(struct ibv_context *context,
                     struct ibv_device_attr *device_attr);

struct ibv_mr *ibv_reg_mr(struct ibv_pd *pd, void *addr,
                          size_t length, int access);


int rdma_bind_addr(struct rdma_cm_id *id, struct sockaddr *addr);

struct rdma_event_channel *rdma_create_event_channel(void);

int rdma_create_id(struct rdma_event_channel *channel,
                   struct rdma_cm_id **id, void *context,
                   enum rdma_port_space ps);

void rdma_destroy_event_channel(struct rdma_event_channel *channel);

int rdma_destroy_id(struct rdma_cm_id *id);

'''

HEADER = PREFIX + '''\
#ifndef SPEAD2_COMMON_IBV_LOADER_H
#define SPEAD2_COMMON_IBV_LOADER_H

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <spead2/common_features.h>

#if SPEAD2_USE_IBV

#include <infiniband/verbs.h>
#include <rdma/rdma_cma.h>

namespace spead2
{

{% for node in nodes -%}
extern {{ node | ptr | gen }};
{% endfor %}

// Load ibverbs. If it could not be loaded, throws std::system_error
void ibv_loader_init();

} // namespace spead2

#endif // SPEAD2_USE_IBV
#endif // SPEAD2_COMMON_IBV_LOADER_H
'''

CXX = '''\
#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif
#include <spead2/common_features.h>

#if SPEAD2_USE_IBV

#include <spead2/common_ibv_loader.h>
#include <spead2/common_ibv_loader_utils.h>
#include <spead2/common_logging.h>
#include <mutex>
#include <exception>

namespace spead2
{

static std::once_flag init_once;
static std::exception_ptr init_result;

{% for node in nodes %}
{{ node | rename(node.name + '_stub') | gen }}
{
{% for arg in (node | args) %}
    (void) {{ arg }};
{% endfor %}
    ibv_loader_stub(init_result);
}
{% endfor %}

{% for node in nodes %}
{{ node | ptr | gen }} = {{ node.name }}_stub;
{% endfor %}

static void reset_stubs()
{
{% for node in nodes %}
    {{ node.name }} = {{ node.name }}_stub;
{% endfor %}
}

static void init()
{
    try
    {
        dl_handle librdmacm("librdmacm.so.1");
        dl_handle libibverbs("libibverbs.so.1");
{% for node in nodes %}
{% set lib = 'libibverbs' if node.name.startswith('ibv_') else 'librdmacm' %}
        {{ node.name }} = reinterpret_cast<{{ node | ptr(False) | gen }}>(
            {{ lib }}.sym("{{ node.name }}"));
{% endfor %}
        // Prevent the libraries being closed, so that the symbols stay valid
        librdmacm.release();
        libibverbs.release();
    }
    catch (std::system_error &e)
    {
        init_result = std::current_exception();
        reset_stubs();
        log_warning("could not load ibverbs: %s", e.what());
    }
}

void ibv_loader_init()
{
    std::call_once(init_once, init);
    if (init_result)
        std::rethrow_exception(init_result);
}

} // namespace spead2

/* Wrappers in the global namespace. This is needed because ibv_exp_create_qp
 * calls ibv_create_qp, and so we need to provide an implementation.
 */
{% for node in nodes if node.name in ['ibv_create_qp'] %}
{{ node | gen }}
{
    return spead2::{{ node.name }}({{ node | args | join(', ') }});
}
{% endfor %}

#endif // SPEAD2_USE_IBV
'''


class RenameVisitor(c_ast.NodeVisitor):
    """Renames a function in the AST.

    The function name is stored both in the Decl and in the inner-most TypeDecl
    of the return type. This only handles the latter.
    """
    def __init__(self, old_name, new_name):
        super(RenameVisitor, self).__init__()
        self.old_name = old_name
        self.new_name = new_name

    def visit_TypeDecl(self, node):
        if node.declname == self.old_name:
            node.declname = self.new_name


def rename_func(func, new_name):
    """Return a copy of a function declaration with a new name"""
    func = copy.deepcopy(func)
    RenameVisitor(func.name, new_name).visit(func)
    func.name = new_name
    return func


def make_func_ptr(func, with_name=True):
    """Create a node of pointer-to-function type."""
    if not with_name:
        node = rename_func(func, None)
    else:
        node = copy.deepcopy(func)
    node.type = c_ast.PtrDecl(quals=[], type=node.type)
    return node


def func_args(func):
    """Get list of function argument names"""
    args = [arg.name for arg in func.type.args.params]
    # Handle (void)
    if args == [None]:
        args = []
    return args


class Visitor(c_ast.NodeVisitor):
    """Collects all the function definitions"""
    def __init__(self):
        self.nodes = []

    def visit_Decl(self, node):
        if not isinstance(node.type, c_ast.FuncDecl):
            return
        self.nodes.append(node)


def gen_node(node):
    return CGenerator().visit(node)


def main(argv):
    environment = jinja2.Environment(autoescape=False, trim_blocks=True)
    environment.filters['gen'] = gen_node
    environment.filters['rename'] = rename_func
    environment.filters['ptr'] = make_func_ptr
    environment.filters['args'] = func_args
    header = environment.from_string(HEADER)
    cxx = environment.from_string(CXX)

    ast = CParser().parse(INPUT)
    visitor = Visitor()
    visitor.visit(ast)
    header_text = header.render(nodes=visitor.nodes)
    cxx_text = cxx.render(nodes=visitor.nodes)
    if len(argv) != 2 or argv[1] not in {'header', 'cxx'}:
        print('Usage: {} header|cxx'.format(argv[0], file=sys.stderr))
        return 1
    elif argv[1] == 'header':
        print(header_text)
    else:
        print(cxx_text)
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))
